//////////////////////////////////////////////////////////////////////////////
#if DODEBUGPRINT
#ifdef __cplusplus
extern "C"
{
#endif
#include <wdm.h>
#ifdef __cplusplus
}
#endif

#include "debugprint.h"
#include <stdio.h>
#include <string.h>


//////////////////////////////////////////////////////////////////////////////
//	Definitions copied from Wdm.h to make this compile in NT4
//	Cross-fingers - lets hope these definitions do not change

#define nt4_IRP_MJ_POWER                    0x16
#define nt4_IRP_MJ_SYSTEM_CONTROL           0x17
#define nt4_IRP_MJ_PNP                      0x1b

#define nt4_IRP_MN_QUERY_DEVICE_RELATIONS       0x07

#define nt4_IRP_MN_SET_POWER                    0x02
#define nt4_IRP_MN_QUERY_POWER                  0x03

typedef enum nt4__DEVICE_RELATION_TYPE {
    nt4_BusRelations,
    nt4_EjectionRelations,
    nt4_PowerRelations,
    nt4_RemovalRelations,
    nt4_TargetDeviceRelation
} nt4_DEVICE_RELATION_TYPE, *nt4_PDEVICE_RELATION_TYPE;

typedef struct _nt4_QueryDeviceRelations
{
    nt4_DEVICE_RELATION_TYPE Type;
} nt4_QueryDeviceRelations, *nt4_PQueryDeviceRelations;

typedef enum _nt4_SYSTEM_POWER_STATE {
    nt4_PowerSystemUnspecified = 0,
    nt4_PowerSystemWorking,
    nt4_PowerSystemSleeping1,
    nt4_PowerSystemSleeping2,
    nt4_PowerSystemSleeping3,
    nt4_PowerSystemHibernate,
    nt4_PowerSystemShutdown,
    nt4_PowerSystemMaximum
} nt4_SYSTEM_POWER_STATE, *nt4_PSYSTEM_POWER_STATE;

typedef enum {
    nt4_PowerActionNone,
    nt4_PowerActionReserved,
    nt4_PowerActionSleep,
    nt4_PowerActionHibernate,
    nt4_PowerActionShutdown,
    nt4_PowerActionShutdownReset,
    nt4_PowerActionShutdownOff
} nt4_POWER_ACTION, *nt4_PPOWER_ACTION;

typedef enum _nt4_DEVICE_POWER_STATE {
    nt4_PowerDeviceUnspecified = 0,
    nt4_PowerDeviceD0,
    nt4_PowerDeviceD1,
    nt4_PowerDeviceD2,
    nt4_PowerDeviceD3,
    nt4_PowerDeviceMaximum
} nt4_DEVICE_POWER_STATE, *nt4_PDEVICE_POWER_STATE;

typedef enum _nt4_POWER_STATE_TYPE {
    nt4_SystemPowerState,
    nt4_DevicePowerState
} nt4_POWER_STATE_TYPE, *nt4_PPOWER_STATE_TYPE;

typedef union _nt4_POWER_STATE {
    nt4_SYSTEM_POWER_STATE SystemState;
    nt4_DEVICE_POWER_STATE DeviceState;
} nt4_POWER_STATE, *nt4_PPOWER_STATE;

typedef struct _nt4_Power {
    ULONG SystemContext;
    nt4_POWER_STATE_TYPE Type;
    nt4_POWER_STATE State;
    nt4_POWER_ACTION ShutdownType;
} nt4_Power, *nt4_PPower;

static BOOLEAN  gbDbgPrintStarted = FALSE;
static char     *gszDbgBuffer = NULL;
#define MAX_DEBUG_BUFFER_SIZE       256
//
//  DebugPrintInit:     Initialize DebugPrint
//
void DebugPrintInit(void)
{
    gszDbgBuffer = (char*)ExAllocatePool(NonPagedPool,MAX_DEBUG_BUFFER_SIZE);
    if(gszDbgBuffer) gbDbgPrintStarted = TRUE;
}

//
//  DebugPrintClose:    Close DebugPrint
//
void DebugPrintClose()
{
    if( !gbDbgPrintStarted) return;
    gbDbgPrintStarted = FALSE;
    if( gszDbgBuffer != NULL)
        ExFreePool(gszDbgBuffer);
}

static char* IrpMajorFunctionNames[] =
{
	"IRP_MJ_CREATE",
	"IRP_MJ_CREATE_NAMED_PIPE",
	"IRP_MJ_CLOSE",
	"IRP_MJ_READ",
	"IRP_MJ_WRITE",
	"IRP_MJ_QUERY_INFORMATION",
	"IRP_MJ_SET_INFORMATION",
	"IRP_MJ_QUERY_EA",
	"IRP_MJ_SET_EA",
	"IRP_MJ_FLUSH_BUFFERS",
	"IRP_MJ_QUERY_VOLUME_INFORMATION",
	"IRP_MJ_SET_VOLUME_INFORMATION",
	"IRP_MJ_DIRECTORY_CONTROL",
	"IRP_MJ_FILE_SYSTEM_CONTROL",
	"IRP_MJ_DEVICE_CONTROL",
	"IRP_MJ_INTERNAL_DEVICE_CONTROL",
	"IRP_MJ_SHUTDOWN",
	"IRP_MJ_LOCK_CONTROL",
	"IRP_MJ_CLEANUP",
	"IRP_MJ_CREATE_MAILSLOT",
	"IRP_MJ_QUERY_SECURITY",
	"IRP_MJ_SET_SECURITY",
	"IRP_MJ_POWER",
	"IRP_MJ_SYSTEM_CONTROL",
	"IRP_MJ_DEVICE_CHANGE",
	"IRP_MJ_QUERY_QUOTA",
	"IRP_MJ_SET_QUOTA",
	"IRP_MJ_PNP",
};
static ULONG n_IrpMajorFunctionNames = sizeof(IrpMajorFunctionNames)/sizeof(char*);

static char* PnPIrpMinorFunctionNames[] =
{
	"IRP_MN_START_DEVICE",
	"IRP_MN_QUERY_REMOVE_DEVICE",
	"IRP_MN_REMOVE_DEVICE",
	"IRP_MN_CANCEL_REMOVE_DEVICE",
	"IRP_MN_STOP_DEVICE",
	"IRP_MN_QUERY_STOP_DEVICE",
	"IRP_MN_CANCEL_STOP_DEVICE",
	"IRP_MN_QUERY_DEVICE_RELATIONS",
	"IRP_MN_QUERY_INTERFACE",
	"IRP_MN_QUERY_CAPABILITIES",
	"IRP_MN_QUERY_RESOURCES",
	"IRP_MN_QUERY_RESOURCE_REQUIREMENTS",
	"IRP_MN_QUERY_DEVICE_TEXT",
	"IRP_MN_FILTER_RESOURCE_REQUIREMENTS",
    "----",
	"IRP_MN_READ_CONFIG",
	"IRP_MN_WRITE_CONFIG",
	"IRP_MN_EJECT",
	"IRP_MN_SET_LOCK",
	"IRP_MN_QUERY_ID",
	"IRP_MN_QUERY_PNP_DEVICE_STATE",
	"IRP_MN_QUERY_BUS_INFORMATION",
	"IRP_MN_DEVICE_USAGE_NOTIFICATION",
	"IRP_MN_SURPRISE_REMOVAL",
	"IRP_MN_QUERY_LEGACY_BUS_INFORMATION",
};
static ULONG n_PnPIrpMinorFunctionNames = sizeof(PnPIrpMinorFunctionNames)/sizeof(char*);

static char* PowerIrpMinorFunctionNames[] =
{
	"IRP_MN_WAIT_WAKE",
	"IRP_MN_POWER_SEQUENCE",
	"IRP_MN_SET_POWER",
	"IRP_MN_QUERY_POWER",
};
static ULONG n_PowerIrpMinorFunctionNames = sizeof(PowerIrpMinorFunctionNames)/sizeof(char*);

static char* WMIIrpMinorFunctionNames[] =
{
	"IRP_MN_QUERY_ALL_DATA",
	"IRP_MN_QUERY_SINGLE_INSTANCE",
	"IRP_MN_CHANGE_SINGLE_INSTANCE",
	"IRP_MN_CHANGE_SINGLE_ITEM",
	"IRP_MN_ENABLE_EVENTS",
	"IRP_MN_DISABLE_EVENTS",
	"IRP_MN_ENABLE_COLLECTION",
	"IRP_MN_DISABLE_COLLECTION",
	"IRP_MN_REGINFO",
	"IRP_MN_EXECUTE_METHOD",
};
static ULONG n_WMIIrpMinorFunctionNames = sizeof(WMIIrpMinorFunctionNames)/sizeof(char*);

static char* PowerSystemStates[] =
{
    "PowerSystemUnspecified",
    "PowerSystemWorking",
    "PowerSystemSleeping1",
    "PowerSystemSleeping2",
    "PowerSystemSleeping3",
    "PowerSystemHibernate",
    "PowerSystemShutdown",
};
static int n_PowerSystemStates = sizeof(PowerSystemStates)/sizeof(char*);

static char* PowerDeviceStates[] =
{
    "PowerDeviceUnspecified",
    "PowerDeviceD0",
    "PowerDeviceD1",
    "PowerDeviceD2",
    "PowerDeviceD3",
};
static int n_PowerDeviceStates = sizeof(PowerDeviceStates)/sizeof(char*);

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
void DbgPrintIrp(PIRP   Irp)
{
    PIO_STACK_LOCATION  IrpStack = IoGetCurrentIrpStackLocation(Irp);
    ULONG   MajorFunction = IrpStack->MajorFunction;
    ULONG   MinorFunction = IrpStack->MinorFunction;
    char    *stkUnknown   = "-Unknown-";

    char    *MajorFnName = NULL;
    char    *MinorFnName = NULL;
    if( !gbDbgPrintStarted) return;
    MajorFnName = (MajorFunction < n_IrpMajorFunctionNames) ? IrpMajorFunctionNames[MajorFunction] : stkUnknown;
    switch (MajorFunction)
    {
        case    nt4_IRP_MJ_PNP:
            MinorFnName = (MinorFunction < n_PnPIrpMinorFunctionNames) ? PnPIrpMinorFunctionNames[MinorFunction] : stkUnknown;
            break;

        case    nt4_IRP_MJ_POWER:
            MinorFnName = (MinorFunction<n_PowerIrpMinorFunctionNames) ? PowerIrpMinorFunctionNames[MinorFunction] : stkUnknown;
            break;

        case    nt4_IRP_MJ_SYSTEM_CONTROL:
            MinorFnName = (MinorFunction<n_WMIIrpMinorFunctionNames) ? WMIIrpMinorFunctionNames[MinorFunction] : stkUnknown;
            
        default:
            MinorFnName = stkUnknown;
    }
    sprintf(gszDbgBuffer,"Irp: MajorFn: %s(%x), MinorFn: %s(%x)\n",
               MajorFnName, MajorFunction, MinorFnName, MinorFunction);

    DbgPrint(gszDbgBuffer);
    return;
}

#if 0
BOOLEAN PrintIrp( char* buffer, int max, int* pbufpos, PIRP Irp)
{
	PIO_STACK_LOCATION IrpStack = IoGetCurrentIrpStackLocation(Irp);
	ULONG MajorFunction = IrpStack->MajorFunction;
	ULONG MinorFunction = IrpStack->MinorFunction;
    if( MajorFunction<n_IrpMajorFunctionNames)
	{
		char* MajorFunctionName = IrpMajorFunctionNames[MajorFunction];
		char* MinorFunctionName = NULL;
		char* ExtraDetails1 = NULL;
		char* ExtraDetails2 = NULL;

		if( PrintANSIString(buffer,max,pbufpos,MajorFunctionName,0)) return TRUE;

		switch(MajorFunction)
		{
        case nt4_IRP_MJ_PNP:
            if( MinorFunction<n_PnPIrpMinorFunctionNames)
				MinorFunctionName = PnPIrpMinorFunctionNames[MinorFunction];
			else
			{
				PHDput(':');
				if( PrintULONGhex(buffer,max,pbufpos,MinorFunction,8)) return TRUE;
			}
            if( MinorFunction==nt4_IRP_MN_QUERY_DEVICE_RELATIONS)
			{
                nt4_PQueryDeviceRelations pQueryDeviceRelations= (nt4_PQueryDeviceRelations)&IrpStack->Parameters.Read.Length;
				switch( pQueryDeviceRelations->Type)
				{
                case nt4_BusRelations:          ExtraDetails1 = "BusRelations"; break;
                case nt4_EjectionRelations:     ExtraDetails1 = "EjectionRelations"; break;
                case nt4_PowerRelations:        ExtraDetails1 = "PowerRelations"; break;
                case nt4_RemovalRelations:      ExtraDetails1 = "RemovalRelations"; break;
                case nt4_TargetDeviceRelation:  ExtraDetails1 = "TargetDeviceRelation"; break;
				}
			}
			break;
        case nt4_IRP_MJ_POWER:
            if( MinorFunction<n_PowerIrpMinorFunctionNames)
				MinorFunctionName = PowerIrpMinorFunctionNames[MinorFunction];
            if( MinorFunction==nt4_IRP_MN_SET_POWER || MinorFunction==nt4_IRP_MN_QUERY_POWER)
			{
                nt4_PPower pPower = (nt4_PPower)&IrpStack->Parameters.Read.Length;
                nt4_POWER_STATE PowerState = pPower->State;
                ExtraDetails1 = (MinorFunction==nt4_IRP_MN_SET_POWER ? "Set Power" : "Query Power");
                if( pPower->Type==nt4_SystemPowerState)
				{
                    if( PowerState.SystemState<n_PowerSystemStates)
						ExtraDetails2 = PowerSystemStates[PowerState.SystemState];
				}
				else
				{
                    if( PowerState.DeviceState<n_PowerDeviceStates)
						ExtraDetails2 = PowerDeviceStates[PowerState.DeviceState];
				}
			}
			break;
        case nt4_IRP_MJ_SYSTEM_CONTROL:
            if( MinorFunction<n_WMIIrpMinorFunctionNames)
				MinorFunctionName = WMIIrpMinorFunctionNames[MinorFunction];
			break;
		}
		if( MinorFunctionName!=NULL)
		{
			PHDput(':');
			if( PrintANSIString(buffer,max,pbufpos,MinorFunctionName,0)) return TRUE;
		}
		if( ExtraDetails1!=NULL)
		{
			PHDput(' ');
			if( PrintANSIString(buffer,max,pbufpos,ExtraDetails1,0)) return TRUE;
		}
		if( ExtraDetails2!=NULL)
		{
			PHDput(' ');
			if( PrintANSIString(buffer,max,pbufpos,ExtraDetails2,0)) return TRUE;
		}
	}
	return FALSE;
}
#endif

//
// DbgDumpData:  Display a series of hex bytes of data upto the length specified.
//
void DbgDumpData(char *signon, unsigned char *data, int len)
{
    int i;
    char szTmpBuffer[3];

    if( !gbDbgPrintStarted) return;
    strcpy(gszDbgBuffer,signon);
    for(i=0;i<len;i++)
    {
        sprintf(szTmpBuffer, "%02X ",data[i]);
        strcat(gszDbgBuffer,szTmpBuffer);
        if (!(i % 15) && i)
        {
            strcat(gszDbgBuffer,"\n");
            strcat(gszDbgBuffer,signon);
            continue;
        }
        if (!(i % 7) && i && ((i % 7) & 1))
        {
           strcat(gszDbgBuffer, "- ");
        }
    }
    if (i % 20)
    {
        strcat(gszDbgBuffer,"\n");
    }
    DbgPrint(gszDbgBuffer);
}

#endif  // DODEBUGPRINT
