#ifndef _SMBIOS_H_
#define _SMBIOS_H_


#include "ntddk.h"

#define SMBIOS_STRING			"_SM_"
#define SMBIOS_MEM_RANGE_BASE       0xF0000
#define SMBIOS_MAPPED_RANGE_LENGTH  0xFFFE
#define SMBIOS_MAJOR_REVISION		2
#define SMBIOS_MINOR_REVISION		3
#define BYTE_BOUNDARY_64K           0xFFF

#define IPMI_DEVICE_INFORMATION     38  // Record Type # 38 in SMBIOS
#ifndef BYTE
typedef unsigned char   BYTE;
#endif

#ifndef WORD
typedef unsigned short  WORD;
#endif

#ifndef DWORD
typedef unsigned long   DWORD;
#endif

#ifndef LONGLONG
typedef __int64         LONGLONG;
#endif

#ifndef ULONG
typedef unsigned long   ULONG;
#endif
#define IO_MAPPED_ADDRESS       0x01
typedef enum
{
    BYTE_BOUNDARY = 0x00,
    BIT32_BOUNDARY,
    BYTE16_BOUNDARY
};

#pragma pack(1)
typedef struct	tagSMBIOSEntryPoint
{
	char	schAnchorString[4];				// "_SM_"
	BYTE	byCheckSum;						// Checksum of the entrypoint structure. (EPS)
	BYTE	byLength;						// Length of the EPS
	BYTE	bySmBiosMajorVersion;			// Major version of the specification.
	BYTE	bySmBiosMinorVersion;			// Minor version of the specification.
	WORD	wMaxSize;						// Size of the largest SMBIOS structure.
	BYTE	byepRevision;					// Entry point revision.
	BYTE	byFmtArea[5];					// Revision interpretation.
	BYTE	byImAnchor[5];					// "_DMI_"
	BYTE	byImCheckSum;					// Checksum of intermediate EPS
	WORD	wTableLength;					// Total Length of the SMBIOS str. table.
    ULONG   ulTableAddress;                 // 32-bit physical starting address.
	WORD	nSmBiosStructures;				// Total number of structures present.
	BYTE	bybcdRevision;					// BCD value of the version.
} SMBIOSEntryPoint, *PSMBIOSEntryPoint;

typedef	enum	tagInterfaceType
{
	Unknown = 0x00,
	KCS,
	SMIC,
	BT
} InterfaceType, *PInterfaceType;

typedef struct	tadSMBIOSHeader
{
	BYTE	Type;
	BYTE	Length;
	WORD	Handle;
} SMBIOSHeader, *PSMBIOSHeader;

typedef struct	tagIPMIDeviceInformation
{
	SMBIOSHeader	smbhdr;
    BYTE            itInterfaceType;
    BYTE            SpecRevision;
    BYTE            I2CSlaveAddress;
    BYTE            NVDevAddress;
    ULONGLONG       BaseAddress;
    BYTE            bBamIntrTrigger     :   1;
    BYTE            bBamIntrPolarity    :   1;
    BYTE            bBamIntrReserved    :   1;
    BYTE            bBamInterruptInfo   :   1;
    BYTE            bBamLSBitAddress    :   1;
    BYTE            bBamReserved        :   1;
    BYTE            bBamRegisterSpacing :   2;
    BYTE            InterruptNumber;        //extension of 2.3.1 Spec.
} IPMIDeviceInformation, *PIPMIDeviceInformation;	

typedef struct  _DEVICE_EXTENSION *PDEVICE_EXTENSION;
#pragma pack()
NTSTATUS
GetIPMIDeviceInformation (
	IN OUT	PDEVICE_EXTENSION	pDeviceExtension
	);

#endif
