/*M*                     
//                    INTEL CORPORATION PROPRIETARY INFORMATION
//      This software is supplied under the terms of a license agreement or
//      nondisclosure agreement with Intel Corporation and may not be copied
//      or disclosed except in accordance with the terms of that agreement.
//            Copyright (c) 1999 Intel Corporation.  All Rights Reserved.
//
*M*/
/////////////////////////////////////////////////////////
// Name:    SMBiosDefs.h
// Module:	PMAL
// Purpose: Contains the SMBIOS specific defines used by the
// CProcessor.cpp.
/////////////////////////////////////////////////////////
//   PVCS Information:
//       $Workfile:   mi6iv61334.tmp  $
//       $Revision:   1.0  $
//       $Modtime:   Dec 14 2001 11:16:14  $
//       $Author:   jabays  $
//	  Revision History:
/*  $Log:   //epg-scc/sst/ProjectDatabases/WMI_PI/archives/ISM6_1/Source/imbdrv/SMBiosDefs.h.v  $
 * 
 *    Rev 1.0   Apr 22 2003 11:14:42   jabays
 * Initial revision.
 * 
 *    Rev 1.1   18 Sep 2002 13:02:44   hnagaraX
 * Added PSMBIOS_SYSTEM_INFORMATION definition
 * 
 *    Rev 1.2   14 Dec 2001 11:22:22   mrajmohx
 * Added code changes to implement the IA_BIOS class in the PI.
 * 
 * COde changes Acked by Jamie
 * 
 *    Rev 1.1   Oct 20 2000 09:30:12   mrajmohx
 * Version control and comments added.
*/

#ifndef _SMBIOS_DEFS_H_
#define _SMBIOS_DEFS_H_
	

#define SMBIOS_ANCHOR_STRING		"_SM_"
#define SMBIOS_MEM_RANGE_BASE       0xF0000
#define SMBIOS_MAPPED_RANGE_LENGTH  0xFFFE
#define SMBIOS_MAJOR_REVISION		2
#define SMBIOS_MINOR_REVISION		3
#define BYTE_BOUNDARY_64K           0xFFF


#pragma pack(1)
typedef struct	_SMBIOS_ENTRY_POINT_STRUCTURE_ {
	char	schAnchorString[4];				// "_SM_"
	BYTE	byCheckSum;						// Checksum of the entrypoint structure. (EPS)
	BYTE	byLength;						// Length of the EPS
	BYTE	bySmBiosMajorVersion;			// Major version of the specification.
	BYTE	bySmBiosMinorVersion;			// Minor version of the specification.
	WORD	wMaxSize;						// Size of the largest SMBIOS structure.
	BYTE	byepRevision;					// Entry point revision.
	BYTE	byFmtArea[5];					// Revision interpretation.
	BYTE	byImAnchor[5];					// "_DMI_"
	BYTE	byImCheckSum;					// Checksum of intermediate EPS
	WORD	wTableLength;					// Total Length of the SMBIOS str. table.
    ULONG   ulTableAddress;                 // 32-bit physical starting address.
	WORD	nSmBiosStructures;				// Total number of structures present.
	BYTE	bybcdRevision;					// BCD value of the version.
} SMBIOS_ENTRY_POINT_STRUCTURE, *PSMBIOS_ENTRY_POINT_STRUCTURE;


typedef struct	_SMBIOS_HEADER_ {
	BYTE	Type;
	BYTE	Length;
	WORD	Handle;
} SMBIOS_HEADER, *PSMBIOS_HEADER;


typedef struct _SMBIOS_PROCESSOR_INFO_ {
	SMBIOS_HEADER	smBiosHeader;
	BYTE	SocketDesignation;
	BYTE	ProcessorType;
	BYTE	ProcessorFamily;
	BYTE	ProcessorManufacturer;
	LARGE_INTEGER ProcessorId;
	BYTE	ProcessorVersion;
	BYTE	Voltage;
	WORD	ExternalClock;
	WORD	MaxSpeed;
	WORD	CurrentSpeed;
	BYTE	Status;
	BYTE	ProcessorUpgrade;
	WORD	L1CacheHandle;
	WORD	L2CacheHandle;
	WORD	L3CacheHandle;
} SMBIOS_PROCESSOR_INFO, *PSMBIOS_PROCESSOR_INFO;
					      


// SMBIOS Structure Type defines
#define SMB_STRUCTURE_TYPE_BIOS_INFO						0x00	// type 0
#define SMB_STRUCTURE_TYPE_SYSTEM_INFO						0x01	// type 1
#define SMB_STRUCTURE_TYPE_SYSTEM_ENCLOSURE					0x03	// type 3
#define SMB_STRUCTURE_TYPE_PROCESSOR_INFO					0x04	// type 4
#define SMB_STRUCTURE_TYPE_CACHE_INFO						0x07	// type 7
#define SMB_STRUCTURE_TYPE_SYSTEM_SLOTS						0x09	// type 9
#define SMB_STRUCTURE_TYPE_PHYSICAL_MEMORY_ARRAY			0x10	// type 16
#define SMB_STRUCTURE_TYPE_MEMORY_DEVICE					0x11	// type 17
#define SMB_STRUCTURE_TYPE_MEMORY_ARRAY_MAPPED_ADDRESS		0x13	// type 19
#define SMB_STRUCTURE_TYPE_MEMORY_DEVICE_MAPPED_ADDRESS		0x14	// type 20
#define SMB_STRUCTURE_TYPE_SYSTEM_BOOT_INFO					0x20	// type 32

// SMBIOS Processor Information specific define
#define SMB_PROC_INFO_CPU_SOCKET_POPULATED					0x40

typedef struct _SMBIOS_SYSTEM_INFORMATION_ {
	SMBIOS_HEADER	SMBIOSHeader;
	BYTE			Manufacturer;
	BYTE			ProductName;
	BYTE			Version;
	BYTE			SerialNumber;
	UUID			Uuid;
	BYTE			WakeupType;
} SMBIOS_SYSTEM_INFORMATOIN, *PSMBIOS_SYSTEM_INFORMATION;

#endif // _SMBIOS_DEFS_H_
