#include "hist.h"
#include "wdm.h"
#include "DiagMsg.h"
#include "DiagCodes.h"
#include "Debugprint.h"

#ifdef DO_HIST

#define MAX_HIST        500

static DWORD ImbDelayHist[MAX_HIST];
// If time based profiling is required, enable the following code.
// However, there will be some performance hit due to QueryPerfCounter call.

#if 0
static LARGE_INTEGER    qHistStart, qHistStop, qHistFreq;
static LARGE_INTEGER    qAveTime = {0}, qNoTimes = {0};
#endif

void InitializeHist(void)
{
   DWORD i; 
   //--------------------------------------------------------------------------
   // Init the delay histogram array
   //
    for(i=0;i<MAX_HIST;i++)
    {
        ImbDelayHist[i] = 0;
    }

}
#if 0
void StartHist(void)
{
    qHistStart = KeQueryPerformanceCounter(NULL);
}
#endif

void UpdateHist(DWORD dwCurValue)
{
#if 0
    qHistStop = KeQueryPerformanceCounter(&qHistFreq);
    qAveTime.QuadPart *= (qNoTimes.QuadPart);
    qAveTime.QuadPart += (qHistStop.QuadPart - qHistStart.QuadPart);
    qAveTime.QuadPart /= (++qNoTimes.QuadPart);
#endif
    if (dwCurValue >= MAX_HIST)
    {
        dwCurValue = MAX_HIST - 1;  // saturate the count.
    }
    InterlockedIncrement(&ImbDelayHist[dwCurValue]);

}

void DisplayHist(void)
{
    DWORD   dwTotalHits, dwMaxHits, dwSlotIndex;
    DWORD   dwSecondMax, dwSecondSlot;
    DWORD   i;
    for(i=dwTotalHits=dwMaxHits=dwSlotIndex=0; i < MAX_HIST;i++)
    {
        dwTotalHits += ImbDelayHist[i];
        if (dwMaxHits < ImbDelayHist[i])
        {
            dwSecondMax = dwMaxHits;
            dwSecondSlot = dwSlotIndex;
            dwMaxHits = ImbDelayHist[i];
            dwSlotIndex = i;
        }
    }
    D_DiagMsg3(D_HighLevelInfo, "Hist. Total=%ld, Max Hits(%ld) at %ld slot",
                    dwTotalHits, dwMaxHits, dwSlotIndex);
    D_DiagMsg3(D_HighLevelInfo, "Hist. Total=%ld, 2nd Max Hits(%ld) at %ld slot",
                    dwTotalHits, dwSecondMax, dwSecondSlot);

#if 0
    D_DiagMsg2(D_HighLevelInfo, "Average Time Taken = %ld us, Total Hits: %ld", ((qAveTime.QuadPart * 1000000)/qHistFreq.QuadPart), qNoTimes.QuadPart);
#endif
}

// This function is called during driver unload to just print all the
// data held in the array to the debugger. That data can be cut and pasted
// on a excel spreadsheet to provide the histogram chart.
void DisplayCompleteHistDatabase(void)
{
    DWORD   i;
    D_DiagMsg0(D_HighLevelInfo, "------ HISTOGRAM DATABASE ------");
    DisplayHist();
    for(i=0;i<MAX_HIST;i++)
    {
        DbgPrint("%ld\n",ImbDelayHist[i]);
    }
}

#endif
