//*****************************************************************************
//* FILE NAME:       w2k_drv.h
//* FILE CONTAINS:
//* DESCRIPTION:     Win 2000 Driver provides standard PnP, PM and WMI support
//*                  for Win 2000.
//* NOTES:           This file is used for Win 2000 only.
//*****************************************************************************
//*                   INTEL CORPORATION PROPRIETARY INFORMATION
//*     This software is supplied under the terms of a license agreement or
//*     nondisclosure agreement with Intel Corporation and may not be copied
//*     or disclosed except in accordance with the terms of that agreement.
//*           Copyright (c) 1995 Intel Corporation.  All Rights Reserved.
//*****************************************************************************

#ifndef D_w2k_drv_HeaderFile
#define D_w2k_drv_HeaderFile

#include "CompDiag.h"
#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("Include")
#endif

//*****************************************************************************
//* INCLUDES
//*****************************************************************************

//-----------------------------------------------------------------------------
// Need to order "ntddk.h" before "windef.h"
//-----------------------------------------------------------------------------

#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("#include <ntddk.h>")
#endif

#include <ntddk.h>

#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("#include <windef.h>")
#endif

#include <windef.h>

#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("#include <wdmguid.h>")
#endif

#include <wdmguid.h>

#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("#include <wmilib.h>")
#endif

#include <wmilib.h>

#ifdef D_CompilerDiagnostics
   #pragma D_CompDiagMsg("#include <wmistr.h>")
#endif

#include <wmistr.h>

//*****************************************************************************
//* DEFINES
//*****************************************************************************

//-----------------------------------------------------------------------------
// Default Maximum Numbre of Devices that can be added for a driver, i.e. the
// number of time that "AddDevice(...)" can be called.
//-----------------------------------------------------------------------------

#define D_DefaultMaxNumberOfDevices  1

//*****************************************************************************
//* ENUMs
//*****************************************************************************

//-----------------------------------------------------------------------------
// Debug Level
//-----------------------------------------------------------------------------

enum E_DebugLevel
  {
   E_DebugLevel_NULL = 0x00000000
   }; 

//-----------------------------------------------------------------------------
// Driver Configuration
//
// Bits specified by driver "Configuration" registry key.
//-----------------------------------------------------------------------------

enum E_DriverConfiguration
  {
   E_DrvConf_NULL = 0x00000000L,

   //--------------------------------------------------------------------------
   // Configure as a Win 2000 driver, i.e. not legacy driver.
   //--------------------------------------------------------------------------

   E_DrvConf_Win2000Driver = 0x00000001L,

   //--------------------------------------------------------------------------
   // Use WMI interface to replace IOCTL interface.
   //--------------------------------------------------------------------------

   E_DrvConf_WMI_Interface = 0x00000002L,

   //--------------------------------------------------------------------------
   // Disable legacy interface to use symbolic link name for backward
   // compatibility; although, the driver can not be unloaded on a Win 2000
   // platform.  This interface can co-exist with the GUID interface.
   //--------------------------------------------------------------------------

   E_DrvConf_DisableLegacyInterface = 0x00000004L,

   //--------------------------------------------------------------------------
   // Enable GUID interface to use GUID to connect to driver.  This interface
   // can co-exist with the legacy interface.
   //--------------------------------------------------------------------------

   E_DrvConf_Unused_2 = 0x00000008L,

   //--------------------------------------------------------------------------
   // If unable to determine if platform is Win 2000 assume Win 2000; otherwise,
   // assume NT 4.0.
   //--------------------------------------------------------------------------

   E_DrvConf_DefaultToWin2000 = 0x00000010L
   };

//=============================================================================
// PnP Device States
//=============================================================================

enum E_PnP_State
  {
   E_PnP_State_DeviceNotPresent,
   E_PnP_State_Stopped,
   E_PnP_State_Started,
   E_PnP_State_StopPending,
   E_PnP_State_RemovePending,
   E_PnP_State_SuprisedRemoved
   };

//*****************************************************************************
//* STRUCTs
//*****************************************************************************

//-----------------------------------------------------------------------------
// Default Win 2000 Device Extension
//
// Device extension for device object created via "IoCreateDevice(...)".  This
// contains the minimum information for Win 2000 specific operations.  If a
// device extension strucure already exists, then this default must be made as
// a substructure.
//
// Members:
//
//    Lower Device Object - The device object below the device object created
//    by "IoCreateDevice(...)".  Usually the device object created by
//    "IoCreateDevice(...)" is the FDO and the lower device objet is the PDO.
//
//    PnP Device State - The PnP state of the FDO.
//-----------------------------------------------------------------------------

struct S_Win2000_DevExt
  {
   PDRIVER_OBJECT       tpDriverObj;
   PDEVICE_OBJECT       tpLowerDeviceObject;
   PDEVICE_OBJECT       tpPDO;
   enum E_PnP_State     ePnP_DeviceState;
   BOOL                 bHaveResources;
	LONG                 lOpenHandleCount;	            // Num of "create file"
	LONG                 lPendingIO_Count;		         // Pending I/O's
   BOOL                 bPaused;
   BOOL                 bDeviceConnected;
   BOOL                 bStopping;                    // Stopping?
	KEVENT               t_KEVENT_Stopping;            // Pending I/O complete?
   BOOL                 bRemoveDevice;
	UNICODE_STRING	      tInterfaceSymLinkName;
 	WMILIB_CONTEXT       t_WMILIB_CONTEXT;
   void                *vp_WMI_Storage;               // WMI data
   };

//*****************************************************************************
//* EXTERN VARIABLES
//*****************************************************************************

//-----------------------------------------------------------------------------
// Saved Registry Path
//
// Used to save the registry path provided to "DriverEntry(...)" and to be
// used at some later time.
//
// NOTE: This is also used as the "base name" for WMI instances.
//-----------------------------------------------------------------------------

extern UNICODE_STRING gtSavedRegistryPath;

//-----------------------------------------------------------------------------
// Saved Registry Subkey
//
// Used as the "base name" for WMI instances.
//-----------------------------------------------------------------------------

extern WCHAR *gwcpSavedRegistrySubkey;

//-----------------------------------------------------------------------------
// Interface Class GUID
//
// Points to the class GUID that identifies the functionality (the device
// interface) being registered.  This is provided by "DriverEntry(...)".
//-----------------------------------------------------------------------------

#ifdef D_InitInterfaceClassGUID

   #ifdef D_CompilerDiagnostics
      #pragma D_CompDiagMsg("Initialize interface GUID in this module")
   #endif

   CONST GUID *gtpInterfaceClassGUID = NULL;

#else

   extern CONST GUID *gtpInterfaceClassGUID;

#endif

//-----------------------------------------------------------------------------
// Driver configuration from registry key from INF file.
//-----------------------------------------------------------------------------

extern unsigned long gulDriverConfiguration;

//*****************************************************************************
//* EXTERN FUNCTIONS
//*****************************************************************************

//=============================================================================
// User Provided "extern" functions.  This are functions that must be provided
// by the "DriverEntry(...)" module in order to "hook-up" to the functions
// for Win 2000 support.
//=============================================================================

//-----------------------------------------------------------------------------
// Initialize Driver
//
// This function is defined in same module as "DriverEntry(...)" and provides
// initialization that was previously being performed in "DriverEntry(...)"
// but now moved to "InitializeDriver(...)".
//
// Arguments:
//
//    Driver Object - Specifies same as the driver object passed to 
//    "DriverEntry(...)".
//
//    Device Object - Receives the device object created by
//    "IoCreateDevice(...)".
//
// Returns:
//
//    NT Status.
//-----------------------------------------------------------------------------

extern NTSTATUS gtInitializeDriver( IN PDRIVER_OBJECT    atpDriverObject,
                                    OUT PDEVICE_OBJECT  *atppDeviceObject
                                    );

//-----------------------------------------------------------------------------
// Get Default Win 2000 Device Extention
//
// This returns a pointer to the portion of the device extention that is
// identified as the "default Win 2000 device extention".
//
// Arguments:
//
//    Device Object - Specifies the device object for which to get the device
//    extension.
//
// Returns:
//
//    Pointer to structure of Win 2000 device extension.
//
//-----------------------------------------------------------------------------

extern struct S_Win2000_DevExt *gspGetWin2000_DevExt
        (
         IN PDEVICE_OBJECT atpDevObj
         );

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------

extern VOID gvPrepareToUnload(PDRIVER_OBJECT DriverObject);

//=============================================================================
// These are function(s) that must be called by "DriverEntry(...)" in order to
// "hook-up" to gain Win 2000 support.
//=============================================================================

//=============================================================================
// gbNewDriverConfiguration(...)
//-----------------------------------------------------------------------------
// If new driver configuration, then configure for new style driver as a
// function of the "Configuration" registry key?
//=============================================================================

extern BOOL gbNewDriverConfiguration
  (
   IN PDRIVER_OBJECT       atpDriverObject,
   IN PUNICODE_STRING      atpRegistryPath,
   IN short               *ashpDriverRegistrySubkey,
   unsigned                auMaxNumberOfDevices,
   IN CONST GUID          *atpInterfaceClassGUID,
   WMIGUIDREGINFO         *asa_WMIGUIDREGINFO,
   const unsigned          acuNumWMI_GUIDs
   );

//=============================================================================
// (...)
//-----------------------------------------------------------------------------
//
//=============================================================================

extern NTSTATUS gtCreateFile(IN PDEVICE_OBJECT atpFDO,IN PIRP atpIRP);

//=============================================================================
// (...)
//-----------------------------------------------------------------------------
//
//=============================================================================

extern NTSTATUS gtCloseFile(IN PDEVICE_OBJECT atpFDO,IN PIRP atpIRP);

//=============================================================================
// (...)
//-----------------------------------------------------------------------------
//
//=============================================================================

extern NTSTATUS gtLockDevice(IN PDEVICE_OBJECT atpFDO,IN PIRP atpIRP);

//=============================================================================
// (...)
//-----------------------------------------------------------------------------
//
//=============================================================================

extern NTSTATUS gtUnlockDevice(IN PDEVICE_OBJECT atpFDO,IN PIRP atpIRP);

#endif