//*****************************************************************************
//* FILE NAME:       w2k_pm.c
//* FILE CONTAINS:
//* DESCRIPTION:     PM support for Win 2000.
//* NOTES:           This file is used for Win 2000 only.
//*****************************************************************************
//*                   INTEL CORPORATION PROPRIETARY INFORMATION
//*     This software is supplied under the terms of a license agreement or
//*     nondisclosure agreement with Intel Corporation and may not be copied
//*     or disclosed except in accordance with the terms of that agreement.
//*           Copyright (c) 1995 Intel Corporation.  All Rights Reserved.
//*****************************************************************************

//*****************************************************************************
//* INCLUDES
//*****************************************************************************

//-----------------------------------------------------------------------------
// Local Includes
//-----------------------------------------------------------------------------

#include "CompDiag.h"
#include "w2k_drv.h"
#include "w2k_pm.h"
#include "DiagMsg.h"

//*****************************************************************************
//* FUNCTIONS
//*****************************************************************************

//=============================================================================
// gtDecodePM_IRP(...)
//-----------------------------------------------------------------------------
// Decode PM IRP
//
// Decode the PM minor function IRP.  Called when PM IRP is received.
//
// Arguments:
//
//    Minor Function Code - Specifies the specific minor function code.
//
//    Default Win 2000 Device Extension - This contains the info specific to
//    the Win 2000 implementation.
//
//    IRP pointer - This is the IRP to be decoded.
//
// Returns:
//
//    STATUS_SUCCESS if successful.
//=============================================================================

NTSTATUS gtDecodePM_IRP
  (
   UCHAR                            aucMinorFuncCode,
   struct S_Win2000_DevExt         *astpFDO_DevExt,
   IN PIRP                          atpIRP
   )
  {
   NTSTATUS tStatus = STATUS_SUCCESS;

   //--------------------------------------------------------------------------
   // Decode the PM minor function code.
   //--------------------------------------------------------------------------

   switch (aucMinorFuncCode)
     {
      case IRP_MN_WAIT_WAKE:
      case IRP_MN_POWER_SEQUENCE:
      case IRP_MN_SET_POWER:
      case IRP_MN_QUERY_POWER:

         break;
      }

   //--------------------------------------------------------------------------
   // Pass on IRP to lower-level device object.  The PDO will pass the IRP to
   // The bus driver.
   //--------------------------------------------------------------------------

   PoStartNextPowerIrp(atpIRP);
   IoSkipCurrentIrpStackLocation(atpIRP);

   tStatus = PoCallDriver(astpFDO_DevExt->tpLowerDeviceObject,atpIRP);

   return (tStatus);
   }
